/* Global variables and functions. */
/* #includes */ /*{{{C}}}*//*{{{*/
#undef  _POSIX_SOURCE
#define _POSIX_SOURCE   1
#undef  _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 2

#include "config.h"

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "auto.h"
#include "error.h"
#include "fs.h"
#include "global.h"
#include "var.h"

#ifdef USE_DMALLOC
#include "dmalloc.h"
#endif
/*}}}*/
/* #defines */ /*{{{*/
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
/*}}}*/

extern char **environ;

static long int intValue(struct Auto *stack, int l) /*{{{*/
{
  struct Value value;
  struct Value *arg=Var_value(Auto_local(stack,l),0,(unsigned int*)0,&value);
  assert(arg->type==V_INTEGER);
  return arg->u.integer;
}
/*}}}*/
static double realValue(struct Auto *stack, int l) /*{{{*/
{
  struct Value value;
  struct Value *arg=Var_value(Auto_local(stack,l),0,(unsigned int*)0,&value);
  assert(arg->type==V_REAL);
  return arg->u.real;
}
/*}}}*/
static struct String *stringValue(struct Auto *stack, int l) /*{{{*/
{
  struct Value value;
  struct Value *arg=Var_value(Auto_local(stack,l),0,(unsigned int*)0,&value);
  assert(arg->type==V_STRING);
  return &(arg->u.string);
}
/*}}}*/

static struct Value *instr(struct Value *v, long int start, long int len, struct String *haystack, struct String *needle) /*{{{*/
{
  const char *haystackChars=haystack->character;
  size_t haystackLength=haystack->length;
  const char *needleChars=needle->character;
  size_t needleLength=needle->length;
  int found;

  --start;
  if (start<0) return Value_new_ERROR(v,OUTOFRANGE,"position");
  if (len<0) return Value_new_ERROR(v,OUTOFRANGE,"length");
  if (((size_t)start)>=haystackLength) return Value_new_INTEGER(v,0);
  haystackChars+=start; haystackLength-=start;
  if (haystackLength>len) haystackLength=len;
  found=1+start;
  while (needleLength<=haystackLength)
  {
    if (memcmp(haystackChars,needleChars,needleLength)==0) return Value_new_INTEGER(v,found);
    ++haystackChars; --haystackLength;
    ++found;
  }
  return Value_new_INTEGER(v,0);
}
/*}}}*/
static struct Value *string(struct Value *v, long int len, char c) /*{{{*/
{
  if (len<0) return Value_new_ERROR(v,OUTOFRANGE,"length");

  Value_new_STRING(v);
  String_size(&v->u.string,len);
  if (len) memset(v->u.string.character,c,len);
  return v;
}
/*}}}*/
static struct Value *mid(struct Value *v, struct String *s, long int position, long int length) /*{{{*/
{
  --position;
  if (position<0) return Value_new_ERROR(v,OUTOFRANGE,"position");
  if (length<0) return Value_new_ERROR(v,OUTOFRANGE,"length");
  if (((size_t)position)+length>s->length)
  {
    length=s->length-position;
    if (length<0) length=0;
  }
  Value_new_STRING(v);
  String_size(&v->u.string,length);
  if (length>0) memcpy(v->u.string.character,s->character+position,length);
  return v;
}
/*}}}*/
static struct Value *inkey(struct Value *v, long int timeout) /*{{{*/
{
  int c;

  if ((c=FS_inkeyChar(0,timeout))==-1)
  {
    if (FS_errmsg) return Value_new_ERROR(v,IOERROR,FS_errmsg);
    else return Value_new_STRING(v);
  }
  else
  {
    Value_new_STRING(v);
    String_appendChar(&v->u.string,c);
    return v;
  }
}
/*}}}*/
static struct Value *env(struct Value *v, long int n) /*{{{*/
{
  int i;

  --n;
  if (n<0) return Value_new_ERROR(v,OUTOFRANGE,"variable number");
  for (i=0; environ[i] && i<n; ++i);
  Value_new_STRING(v);
  if (i==n && environ[i]) String_appendChars(&v->u.string,environ[i]);
  return v;
}
/*}}}*/
static struct Value *rnd(struct Value *v, long int x) /*{{{*/
{
  if (x<0) srand(-x);
  if (x==0 || x==1) Value_new_REAL(v,rand()/(double)RAND_MAX);
  else Value_new_REAL(v,rand()%x+1);
  return v;
}
/*}}}*/

static struct Value *fn_abs(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,fabs(realValue(stack,0)));
}
/*}}}*/
static struct Value *fn_asc(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *s=stringValue(stack,0);

  if (s->length==0) return Value_new_ERROR(v,UNDEFINED,"`asc' of empty string");
  return Value_new_INTEGER(v,s->character[0]);
}
/*}}}*/
static struct Value *fn_atn(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,atan(realValue(stack,0)));
}
/*}}}*/
static struct Value *fn_chr(struct Value *v, struct Auto *stack) /*{{{*/
{
  Value_new_STRING(v);
  String_size(&v->u.string,1);
  v->u.string.character[0]=intValue(stack,0);
  return v;
}
/*}}}*/
static struct Value *fn_cint(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_INTEGER(v,realValue(stack,0));
}
/*}}}*/
static struct Value *fn_cos(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,cos(realValue(stack,0)));
}
/*}}}*/
static struct Value *fn_cvi(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *s=stringValue(stack,0);
  long int n=(s->length && s->character[s->length-1]<0) ? -1 : 0;
  int i;
  
  for (i=s->length-1; i>=0; --i) n=(n<<8)|(s->character[i]&0xff);
  return Value_new_INTEGER(v,n);
}
/*}}}*/
static struct Value *fn_cvs(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *s=stringValue(stack,0);
  float n;

  if (s->length!=sizeof(float)) return Value_new_ERROR(v,BADCONVERSION,"floating point");
  memcpy(&n,s->character,sizeof(float));
  return Value_new_REAL(v,(double)n);
}
/*}}}*/
static struct Value *fn_cvd(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *s=stringValue(stack,0);
  double n;

  if (s->length!=sizeof(double)) return Value_new_ERROR(v,BADCONVERSION,"floating point");
  memcpy(&n,s->character,sizeof(double));
  return Value_new_REAL(v,n);
}
/*}}}*/
static struct Value *fn_date(struct Value *v, struct Auto *stack) /*{{{*/
{
  time_t t;
  struct tm *now;

  Value_new_STRING(v);
  String_size(&v->u.string,10);
  time(&t);
  now=localtime(&t);
  sprintf(v->u.string.character,"%02d-%02d-%04d",now->tm_mon+1,now->tm_mday,now->tm_year+1900);
  return v;
}
/*}}}*/
static struct Value *fn_deg(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,realValue(stack,0)*(180.0/M_PI));
}
/*}}}*/
static struct Value *fn_edit(struct Value *v, struct Auto *stack) /*{{{*/
{
  int code;
  char *begin,*end,*rd,*wr;
  char quote;

  code=intValue(stack,1);
  Value_new_STRING(v);
  String_appendString(&v->u.string,stringValue(stack,0));
  begin=rd=wr=v->u.string.character;
  end=rd+v->u.string.length;

  /* 8 - Discard Leading Spaces and Tabs */
  if (code & 8) while (rd<end && (*rd==' ' || *rd=='\t')) ++rd;

  while (rd<end)
  {
    /*  1 - Discard parity bit */
    if (code&1) *rd=*rd&0x7f;

    /* 2 - Discard all spaces and tabs */
    if ((code&2) && (*rd==' ' || *rd=='\t')) { ++rd; continue; }

    /* 4 - Discard all carriage returns, line feeds, form feeds, deletes, escapes, and nulls */
    if ((code&4) && (*rd=='\r' || *rd=='\n' || *rd=='\f' || *rd==127 || *rd==27 || *rd=='\0')) { ++rd; continue; }

    /* 16 - Convert Multiple Spaces and Tabs to one space */
    if ((code&16) && ((*rd==' ') || (*rd=='\t')))
    {
      *wr++=' ';
      while (rd<end && (*rd==' ' || *rd=='\t')) ++rd; 
      continue;
    }

    /* 32 - Convert lower to upper case */
    if ((code&32) && islower(*rd)) { *wr++=toupper(*rd++); continue; }

    /* 64 - Convert brackets to parentheses */
    if (code&64)
    {
      if (*rd=='[') { *wr++='('; ++rd; continue; }
      else if (*rd==']') { *wr++=')'; ++rd; continue; }
    }

    /* 256 - Suppress all editing for characters within quotation marks */
    if ((code&256) && (*rd=='"' || *rd=='\''))
    {
      quote=*rd;
      *wr++=*rd++;
      while (rd<end && *rd!=quote) *wr++=*rd++;
      if (rd<end) { *wr++=*rd++; quote='\0'; }
      continue;
    }

    *wr++=*rd++;
  }

  /*  128 - Discard Trailing Spaces and Tabs */
  if ((code & 128) && wr>begin)
  {
    while (wr>begin && (*(wr-1)=='\0' || *(wr-1)=='\t')) --wr;
  }

  String_size(&v->u.string,wr-begin);
  return v;
}
/*}}}*/
static struct Value *fn_environi(struct Value *v, struct Auto *stack) /*{{{*/
{
  return env(v,intValue(stack,0));
}
/*}}}*/
static struct Value *fn_environd(struct Value *v, struct Auto *stack) /*{{{*/
{
  return env(v,realValue(stack,0)+0.5);
}
/*}}}*/
static struct Value *fn_environs(struct Value *v, struct Auto *stack) /*{{{*/
{
  char *var;

  Value_new_STRING(v);
  if ((var=stringValue(stack,0)->character))
  {
    char *val=getenv(var);

    if (val) String_appendChars(&v->u.string,val);
  }
  return v;
}
/*}}}*/
static struct Value *fn_eof(struct Value *v, struct Auto *stack) /*{{{*/
{
  int e=FS_eof(intValue(stack,0));

  if (e==-1) return Value_new_ERROR(v,IOERROR,FS_errmsg);
  return Value_new_INTEGER(v,e?-1:0);
}
/*}}}*/
static struct Value *fn_erl(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_INTEGER(v,stack->erl);
}
/*}}}*/
static struct Value *fn_err(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_INTEGER(v,stack->err.type==V_NIL?0:stack->err.u.error.code);
}
/*}}}*/
static struct Value *fn_exp(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,exp(realValue(stack,0)));
}
/*}}}*/
static struct Value *fn_false(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_INTEGER(v,0);
}
/*}}}*/
static struct Value *fn_fix(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_INTEGER(v,(long int)realValue(stack,0));
}
/*}}}*/
static struct Value *fn_hex(struct Value *v, struct Auto *stack) /*{{{*/
{
  char buf[sizeof(long int)*2+1];

  sprintf(buf,"%x",intValue(stack,0));
  Value_new_STRING(v);
  String_appendChars(&v->u.string,buf);
  return v;
}
/*}}}*/
static struct Value *fn_int(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_INTEGER(v,floor(realValue(stack,0)));
}
/*}}}*/
static struct Value *fn_inkey(struct Value *v, struct Auto *stack) /*{{{*/
{
  return inkey(v,0);
}
/*}}}*/
static struct Value *fn_inkeyi(struct Value *v, struct Auto *stack) /*{{{*/
{
  return inkey(v,intValue(stack,0)*10);
}
/*}}}*/
static struct Value *fn_inkeyd(struct Value *v, struct Auto *stack) /*{{{*/
{
  return inkey(v,realValue(stack,0)*10+0.5);
}
/*}}}*/
static struct Value *fn_instr2(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *haystack=stringValue(stack,0);

  return instr(v,1,haystack->length,haystack,stringValue(stack,1));
}
/*}}}*/
static struct Value *fn_instr3iss(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *haystack=stringValue(stack,1);

  return instr(v,intValue(stack,0),haystack->length,haystack,stringValue(stack,2));
}
/*}}}*/
static struct Value *fn_instr3ssi(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *haystack=stringValue(stack,0);

  return instr(v,intValue(stack,2),haystack->length,haystack,stringValue(stack,1));
}
/*}}}*/
static struct Value *fn_instr3dss(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *haystack=stringValue(stack,1);

  return instr(v,realValue(stack,0),haystack->length,haystack,stringValue(stack,2));
}
/*}}}*/
static struct Value *fn_instr3ssd(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *haystack=stringValue(stack,0);

  return instr(v,realValue(stack,2),haystack->length,haystack,stringValue(stack,1));
}
/*}}}*/
static struct Value *fn_instr4ii(struct Value *v, struct Auto *stack) /*{{{*/
{
  return instr(v,intValue(stack,2),intValue(stack,3),stringValue(stack,0),stringValue(stack,1));
}
/*}}}*/
static struct Value *fn_instr4id(struct Value *v, struct Auto *stack) /*{{{*/
{
  return instr(v,intValue(stack,2),realValue(stack,3),stringValue(stack,0),stringValue(stack,1));
}
/*}}}*/
static struct Value *fn_instr4di(struct Value *v, struct Auto *stack) /*{{{*/
{
  return instr(v,realValue(stack,2),intValue(stack,3),stringValue(stack,0),stringValue(stack,1));
}
/*}}}*/
static struct Value *fn_instr4dd(struct Value *v, struct Auto *stack) /*{{{*/
{
  return instr(v,realValue(stack,2),realValue(stack,3),stringValue(stack,0),stringValue(stack,1));
}
/*}}}*/
static struct Value *fn_lcase(struct Value *v, struct Auto *stack) /*{{{*/
{
  Value_new_STRING(v);
  String_appendString(&v->u.string,stringValue(stack,0));
  String_lcase(&v->u.string);
  return v;
}
/*}}}*/
static struct Value *fn_len(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_INTEGER(v,stringValue(stack,0)->length);
}
/*}}}*/
static struct Value *fn_left(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *s=stringValue(stack,0);
  long int len=intValue(stack,1);
  int left=((size_t)len)<s->length ? len : s->length;

  if (left<0) return Value_new_ERROR(v,OUTOFRANGE,"length");
  Value_new_STRING(v);
  String_size(&v->u.string,left);
  if (left) memcpy(v->u.string.character,s->character,left);
  return v;
}
/*}}}*/
static struct Value *fn_lof(struct Value *v, struct Auto *stack) /*{{{*/
{
  long int l=FS_lof(intValue(stack,0));

  if (l==-1) return Value_new_ERROR(v,IOERROR,FS_errmsg);
  return Value_new_INTEGER(v,l);
}
/*}}}*/
static struct Value *fn_log(struct Value *v, struct Auto *stack) /*{{{*/
{
  if (realValue(stack,0)<=0.0) Value_new_ERROR(v,UNDEFINED,"Logarithm of negative value");
  else Value_new_REAL(v,log(realValue(stack,0)));
  return v;
}
/*}}}*/
static struct Value *fn_mid2i(struct Value *v, struct Auto *stack) /*{{{*/
{
  return mid(v,stringValue(stack,0),intValue(stack,1),stringValue(stack,0)->length);
}
/*}}}*/
static struct Value *fn_mid2d(struct Value *v, struct Auto *stack) /*{{{*/
{
  return mid(v,stringValue(stack,0),realValue(stack,1),stringValue(stack,0)->length);
}
/*}}}*/
static struct Value *fn_mid3ii(struct Value *v, struct Auto *stack) /*{{{*/
{
  return mid(v,stringValue(stack,0),intValue(stack,1),intValue(stack,2));
}
/*}}}*/
static struct Value *fn_mid3id(struct Value *v, struct Auto *stack) /*{{{*/
{
  return mid(v,stringValue(stack,0),intValue(stack,1),realValue(stack,2)+0.5);
}
/*}}}*/
static struct Value *fn_mid3di(struct Value *v, struct Auto *stack) /*{{{*/
{
  return mid(v,stringValue(stack,0),realValue(stack,1)+0.5,intValue(stack,2));
}
/*}}}*/
static struct Value *fn_mid3dd(struct Value *v, struct Auto *stack) /*{{{*/
{
  return mid(v,stringValue(stack,0),realValue(stack,1)+0.5,realValue(stack,2)+0.5);
}
/*}}}*/
static struct Value *fn_mki(struct Value *v, struct Auto *stack) /*{{{*/
{
  long int x=intValue(stack,0);
  struct String s;
  size_t i;

  Value_new_STRING(v);
  String_size(&v->u.string,sizeof(long int));
  for (i=0; i<sizeof(long int); ++i,x>>=8) v->u.string.character[i]=(x&0xff);
  return v;
}
/*}}}*/
static struct Value *fn_mks(struct Value *v, struct Auto *stack) /*{{{*/
{
  float x=realValue(stack,0);
  struct String s;
  int i;

  Value_new_STRING(v);
  String_size(&v->u.string,sizeof(float));
  memcpy(v->u.string.character,&x,sizeof(float));
  return v;
}
/*}}}*/
static struct Value *fn_mkd(struct Value *v, struct Auto *stack) /*{{{*/
{
  double x=realValue(stack,0);
  struct String s;
  int i;

  Value_new_STRING(v);
  String_size(&v->u.string,sizeof(double));
  memcpy(v->u.string.character,&x,sizeof(double));
  return v;
}
/*}}}*/
static struct Value *fn_oct(struct Value *v, struct Auto *stack) /*{{{*/
{
  char buf[sizeof(long int)*3+1];

  sprintf(buf,"%o",intValue(stack,0));
  Value_new_STRING(v);
  String_appendChars(&v->u.string,buf);
  return v;
}
/*}}}*/
static struct Value *fn_pi(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,M_PI);
}
/*}}}*/
static struct Value *fn_rad(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,(realValue(stack,0)*M_PI)/180.0);
}
/*}}}*/
static struct Value *fn_right(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *s=stringValue(stack,0);
  int len=s->length;
  int right=intValue(stack,1)<len ? intValue(stack,1) : len;
  if (right<0) return Value_new_ERROR(v,OUTOFRANGE,"length");
  Value_new_STRING(v);
  String_size(&v->u.string,right);
  if (right) memcpy(v->u.string.character,s->character+len-right,right);
  return v;
}
/*}}}*/
static struct Value *fn_rnd(struct Value *v, struct Auto *stack) /*{{{*/
{
  return rnd(v,0);
}
/*}}}*/
static struct Value *fn_rndi(struct Value *v, struct Auto *stack) /*{{{*/
{
  return rnd(v,intValue(stack,0));
}
/*}}}*/
static struct Value *fn_rndd(struct Value *v, struct Auto *stack) /*{{{*/
{
  return rnd(v,realValue(stack,0)+0.5);
}
/*}}}*/
static struct Value *fn_sgn(struct Value *v, struct Auto *stack) /*{{{*/
{
  double x=realValue(stack,0);
  return Value_new_INTEGER(v,x<0.0 ? -1 : (x==0.0 ? 0 : 1));
}
/*}}}*/
static struct Value *fn_sin(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,sin(realValue(stack,0)));
}
/*}}}*/
static struct Value *fn_space(struct Value *v, struct Auto *stack) /*{{{*/
{
  long int len=intValue(stack,0);

  if (len<0) return Value_new_ERROR(v,OUTOFRANGE,"length");
  Value_new_STRING(v);
  String_size(&v->u.string,len);
  if (len) memset(v->u.string.character,' ',len);
  return v;
}
/*}}}*/
static struct Value *fn_sqr(struct Value *v, struct Auto *stack) /*{{{*/
{
  if (realValue(stack,0)<0.0) Value_new_ERROR(v,OUTOFRANGE,"Square root argument");
  else Value_new_REAL(v,sqrt(realValue(stack,0)));
  return v;
}
/*}}}*/
static struct Value *fn_str(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct Value value,*arg;
  struct String s;

  arg=Var_value(Auto_local(stack,0),0,(unsigned int*)0,&value);
  assert(arg->type!=V_ERROR);
  String_new(&s);
  Value_toString(arg,&s,' ',-1,0,0,-1,0);
  v->type=V_STRING;
  v->u.string=s;
  return v;
}
/*}}}*/
static struct Value *fn_stringii(struct Value *v, struct Auto *stack) /*{{{*/
{
  return string(v,intValue(stack,0),intValue(stack,1));
}
/*}}}*/
static struct Value *fn_stringid(struct Value *v, struct Auto *stack) /*{{{*/
{
  return string(v,intValue(stack,0),realValue(stack,1)+0.5);
}
/*}}}*/
static struct Value *fn_stringdi(struct Value *v, struct Auto *stack) /*{{{*/
{
  return string(v,realValue(stack,0)+0.5,intValue(stack,1));
}
/*}}}*/
static struct Value *fn_stringdd(struct Value *v, struct Auto *stack) /*{{{*/
{
  return string(v,realValue(stack,0)+0.5,realValue(stack,1)+0.5);
}
/*}}}*/
static struct Value *fn_stringis(struct Value *v, struct Auto *stack) /*{{{*/
{
  if (stringValue(stack,1)->length==0) return Value_new_ERROR(v,UNDEFINED,"`string$' of empty string");

  return string(v,intValue(stack,0),stringValue(stack,1)->character[0]);
}
/*}}}*/
static struct Value *fn_stringds(struct Value *v, struct Auto *stack) /*{{{*/
{
  if (stringValue(stack,1)->length==0) return Value_new_ERROR(v,UNDEFINED,"`string$' of empty string");

  return string(v,realValue(stack,0)+0.5,stringValue(stack,1)->character[0]);
}
/*}}}*/
static struct Value *fn_tan(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_REAL(v,tan(realValue(stack,0)));
}
/*}}}*/
static struct Value *fn_time(struct Value *v, struct Auto *stack) /*{{{*/
{
  time_t t;
  struct tm *now;

  Value_new_STRING(v);
  String_size(&v->u.string,8);
  time(&t);
  now=localtime(&t);
  sprintf(v->u.string.character,0,"%02d:%02d:%02d",now->tm_hour,now->tm_min,now->tm_sec);
  return v;
}
/*}}}*/
static struct Value *fn_timer(struct Value *v, struct Auto *stack) /*{{{*/
{
  time_t t;
  struct tm *l;

  time(&t);
  l=localtime(&t);
  return Value_new_REAL(v,l->tm_hour*3600+l->tm_min*60+l->tm_sec);
}
/*}}}*/
static struct Value *fn_true(struct Value *v, struct Auto *stack) /*{{{*/
{
  return Value_new_INTEGER(v,-1);
}
/*}}}*/
static struct Value *fn_ucase(struct Value *v, struct Auto *stack) /*{{{*/
{
  Value_new_STRING(v);
  String_appendString(&v->u.string,stringValue(stack,0));
  String_ucase(&v->u.string);
  return v;
}
/*}}}*/
static struct Value *fn_val(struct Value *v, struct Auto *stack) /*{{{*/
{
  struct String *s=stringValue(stack,0);
  
  return Value_new_REAL(v,s->character?strtod(s->character,(char**)0):0.0);
}
/*}}}*/

static unsigned int hash(const char *s) /*{{{*/
{
  unsigned int h=0;

  while (*s)
  {
    h=h*256+tolower(*s);
    ++s;
  }
  return h%GLOBAL_HASHSIZE;
}
/*}}}*/
static void builtin(struct Global *this, const char *ident, enum ValueType type, struct Value *(* func)(struct Value *value, struct Auto *stack), int argLength, ...) /*{{{*/
{
  struct GlobalElement **r;
  struct Symbol *s,**sptr;
  int i;
  va_list ap;

  for
  (
    r=&this->table[hash(ident)];
    *r!=(struct GlobalElement*)0 && cistrcmp((*r)->identifier,ident);
    r=&((*r)->next)
  );
  if (*r==(struct GlobalElement*)0)
  {
    *r=malloc(sizeof(struct GlobalElement));
    (*r)->identifier=strcpy(malloc(strlen(ident)+1),ident);
    (*r)->next=(struct GlobalElement*)0;
    s=&(*r)->sym;
  }
  else
  {
    for (sptr=&((*r)->sym.u.sub.u.bltin.next); *sptr; sptr=&((*sptr)->u.sub.u.bltin.next));
    *sptr=s=malloc(sizeof(struct Symbol));
  }
  s->u.sub.u.bltin.next=(struct Symbol*)0;
  s->type=BUILTINFUNCTION;
  s->u.sub.argLength=argLength;
  s->u.sub.argTypes=argLength ? malloc(sizeof(enum ValueType)*argLength) : (enum ValueType*)0;
  s->u.sub.retType=type;
  va_start(ap,argLength);
  for (i=0; i<argLength; ++i)
  {
    s->u.sub.argTypes[i]=va_arg(ap,enum ValueType);
  }
  va_end(ap);
  s->u.sub.u.bltin.call=func;
}
/*}}}*/

struct Global *Global_new(struct Global *this) /*{{{*/
{
  builtin(this,"abs",     V_REAL,   fn_abs,       1,V_REAL);
  builtin(this,"asc",     V_INTEGER,fn_asc,       1,V_STRING);
  builtin(this,"atn",     V_REAL,   fn_atn,       1,V_REAL);
  builtin(this,"chr$",    V_STRING, fn_chr,       1,V_INTEGER);
  builtin(this,"cint",    V_INTEGER,fn_cint,      1,V_REAL);
  builtin(this,"cos",     V_REAL,   fn_cos,       1,V_REAL);
  builtin(this,"cvi",     V_INTEGER,fn_cvi,       1,V_STRING);
  builtin(this,"cvs",     V_REAL,   fn_cvs,       1,V_STRING);
  builtin(this,"cvd",     V_REAL,   fn_cvd,       1,V_STRING);
  builtin(this,"date$",   V_STRING, fn_date,      0);
  builtin(this,"deg",     V_REAL,   fn_deg,       1,V_REAL);
  builtin(this,"edit$",   V_STRING, fn_edit,      2,V_STRING,V_INTEGER);
  builtin(this,"environ$",V_STRING, fn_environi,  1,V_INTEGER);
  builtin(this,"environ$",V_STRING, fn_environd,  1,V_REAL);
  builtin(this,"environ$",V_STRING, fn_environs,  1,V_STRING);
  builtin(this,"eof",     V_INTEGER,fn_eof,       1,V_INTEGER);
  builtin(this,"erl",     V_INTEGER,fn_erl,       0);
  builtin(this,"err",     V_INTEGER,fn_err,       0);
  builtin(this,"exp",     V_REAL,   fn_exp,       1,V_REAL);
  builtin(this,"false",   V_INTEGER,fn_false,     0);
  builtin(this,"fix",     V_INTEGER,fn_fix,       1,V_REAL);
  builtin(this,"hex$",    V_STRING, fn_hex,       1,V_INTEGER);
  builtin(this,"inkey$",  V_STRING, fn_inkey,     0);
  builtin(this,"inkey$",  V_STRING, fn_inkeyi,    1,V_INTEGER);
  builtin(this,"inkey$",  V_STRING, fn_inkeyd,    1,V_REAL);
  builtin(this,"int",     V_INTEGER,fn_int,       1,V_REAL);
  builtin(this,"instr",   V_INTEGER,fn_instr2,    2,V_STRING,V_STRING);
  builtin(this,"instr",   V_INTEGER,fn_instr3iss, 3,V_INTEGER,V_STRING,V_STRING);
  builtin(this,"instr",   V_INTEGER,fn_instr3ssi, 3,V_STRING,V_STRING,V_INTEGER);
  builtin(this,"instr",   V_INTEGER,fn_instr3dss, 3,V_REAL,V_STRING,V_STRING);
  builtin(this,"instr",   V_INTEGER,fn_instr3ssd, 3,V_STRING,V_STRING,V_REAL);
  builtin(this,"instr",   V_INTEGER,fn_instr4ii,  4,V_STRING,V_STRING,V_INTEGER,V_INTEGER);
  builtin(this,"instr",   V_INTEGER,fn_instr4id,  4,V_STRING,V_STRING,V_INTEGER,V_REAL);
  builtin(this,"instr",   V_INTEGER,fn_instr4di,  4,V_STRING,V_STRING,V_REAL,V_INTEGER);
  builtin(this,"instr",   V_INTEGER,fn_instr4dd,  4,V_STRING,V_STRING,V_REAL,V_REAL);
  builtin(this,"lcase$",  V_STRING, fn_lcase,     1,V_STRING);
  builtin(this,"left$",   V_STRING, fn_left,      2,V_STRING,V_INTEGER);
  builtin(this,"len",     V_INTEGER,fn_len,       1,V_STRING);
  builtin(this,"lof",     V_INTEGER,fn_lof,       1,V_INTEGER);
  builtin(this,"log",     V_REAL,   fn_log,       1,V_REAL);
  builtin(this,"mid$",    V_STRING, fn_mid2i,     2,V_STRING,V_INTEGER);
  builtin(this,"mid$",    V_STRING, fn_mid2d,     2,V_STRING,V_REAL);
  builtin(this,"mid$",    V_STRING, fn_mid3ii,    3,V_STRING,V_INTEGER,V_INTEGER);
  builtin(this,"mid$",    V_STRING, fn_mid3id,    3,V_STRING,V_INTEGER,V_REAL);
  builtin(this,"mid$",    V_STRING, fn_mid3di,    3,V_STRING,V_REAL,V_INTEGER);
  builtin(this,"mid$",    V_STRING, fn_mid3dd,    3,V_STRING,V_REAL,V_REAL);
  builtin(this,"mki$",    V_STRING, fn_mki,       1,V_INTEGER);
  builtin(this,"mks$",    V_STRING, fn_mks,       1,V_REAL);
  builtin(this,"mkd$",    V_STRING, fn_mkd,       1,V_REAL);
  builtin(this,"oct$",    V_STRING, fn_oct,       1,V_INTEGER);
  builtin(this,"pi" ,     V_REAL,   fn_pi,        0);
  builtin(this,"rad",     V_REAL,   fn_rad,       1,V_REAL);
  builtin(this,"right$",  V_STRING, fn_right,     2,V_STRING,V_INTEGER);
  builtin(this,"rnd",     V_INTEGER,fn_rnd,       0);
  builtin(this,"rnd",     V_INTEGER,fn_rndd,      1,V_REAL);
  builtin(this,"rnd",     V_INTEGER,fn_rndi,      1,V_INTEGER);
  builtin(this,"sgn",     V_INTEGER,fn_sgn,       1,V_REAL);
  builtin(this,"sin",     V_REAL,   fn_sin,       1,V_REAL);
  builtin(this,"space$",  V_STRING, fn_space,     1,V_INTEGER);
  builtin(this,"sqr",     V_REAL,   fn_sqr,       1,V_REAL);
  builtin(this,"str$",    V_STRING, fn_str,       1,V_REAL);
  builtin(this,"string$", V_STRING, fn_stringii,  2,V_INTEGER,V_INTEGER);
  builtin(this,"string$", V_STRING, fn_stringid,  2,V_INTEGER,V_REAL);
  builtin(this,"string$", V_STRING, fn_stringdi,  2,V_REAL,V_INTEGER);
  builtin(this,"string$", V_STRING, fn_stringdd,  2,V_REAL,V_REAL);
  builtin(this,"string$", V_STRING, fn_stringis,  2,V_INTEGER,V_STRING);
  builtin(this,"string$", V_STRING, fn_stringds,  2,V_REAL,V_STRING);
  builtin(this,"tan",     V_REAL,   fn_tan,       1,V_REAL);
  builtin(this,"time$",   V_STRING, fn_time,      0);
  builtin(this,"timer",   V_REAL,   fn_timer,     0);
  builtin(this,"true",    V_INTEGER,fn_true,      0);
  builtin(this,"ucase$",  V_STRING, fn_ucase,     1,V_STRING);
  builtin(this,"val",     V_REAL,   fn_val,       1,V_STRING);
  return this;
}
/*}}}*/
int Global_find(struct Global *this, struct Identifier *ident, enum NameSpace nameSpace) /*{{{*/
{
  struct GlobalElement **r;

  for
  (
    r=&this->table[hash(ident->name)];
    *r!=(struct GlobalElement*)0 && (((*r)->sym.type==GLOBALVAR && (*r)->nameSpace!=nameSpace) || cistrcmp((*r)->identifier,ident->name));
    r=&((*r)->next)
  );
  if (*r==(struct GlobalElement*)0) return 0;
  ident->sym=&((*r)->sym);
  return 1;
}
/*}}}*/
int Global_variable(struct Global *this, struct Identifier *ident, enum ValueType type, enum NameSpace nameSpace, int redeclare) /*{{{*/
{
  struct GlobalElement **r;

  for
  (
    r=&this->table[hash(ident->name)];
    *r!=(struct GlobalElement*)0 && (((*r)->sym.type==GLOBALVAR && (*r)->nameSpace!=nameSpace) || cistrcmp((*r)->identifier,ident->name));
    r=&((*r)->next)
  );
  if (*r==(struct GlobalElement*)0)
  {
    *r=malloc(sizeof(struct GlobalElement));
    (*r)->identifier=strcpy(malloc(strlen(ident->name)+1),ident->name);
    (*r)->nameSpace=nameSpace;
    (*r)->next=(struct GlobalElement*)0;
    (*r)->sym.type=GLOBALVAR;
    Var_new(&((*r)->sym.u.var),type,0,(unsigned int*)0);
  }
  else if (redeclare) Var_retype(&((*r)->sym.u.var),type);
  switch ((*r)->sym.type)
  {
    case GLOBALVAR:
    {
      ident->sym=&((*r)->sym);
      break;
    }
    case BUILTINFUNCTION:
    {
      return 0;
      break;
    }
    case USERFUNCTION:
    {
      return 0;
      break;
    }
    default: assert(0);
  }
  return 1;
}
/*}}}*/
int Global_function(struct Global *this, struct Identifier *ident, enum ValueType type, struct Pc *deffn, struct Pc *begin, int argLength, enum ValueType *argTypes) /*{{{*/
{
  struct GlobalElement **r;

  for
  (
    r=&this->table[hash(ident->name)];
    *r!=(struct GlobalElement*)0 && cistrcmp((*r)->identifier,ident->name);
    r=&((*r)->next)
  );
  if (*r!=(struct GlobalElement*)0) return 0;
  *r=malloc(sizeof(struct GlobalElement));
  (*r)->identifier=strcpy(malloc(strlen(ident->name)+1),ident->name);
  (*r)->next=(struct GlobalElement*)0;
  (*r)->sym.type=USERFUNCTION;
  (*r)->sym.u.sub.u.def.scope.start=*deffn;
  (*r)->sym.u.sub.u.def.scope.begin=*begin;
  (*r)->sym.u.sub.argLength=argLength;
  (*r)->sym.u.sub.argTypes=argTypes;
  (*r)->sym.u.sub.retType=type;
  (*r)->sym.u.sub.u.def.localLength=0;
  (*r)->sym.u.sub.u.def.localTypes=(enum ValueType*)0;
  ident->sym=&((*r)->sym);
  return 1;
}
/*}}}*/
void Global_endfunction(struct Global *this, struct Identifier *ident, struct Pc *end) /*{{{*/
{
  struct GlobalElement **r;

  for
  (
    r=&this->table[hash(ident->name)];
    *r!=(struct GlobalElement*)0 && cistrcmp((*r)->identifier,ident->name);
    r=&((*r)->next)
  );
  assert(*r!=(struct GlobalElement*)0);
  (*r)->sym.u.sub.u.def.scope.end=*end;
}
/*}}}*/
void Global_clear(struct Global *this) /*{{{*/
{
  int i;

  for (i=0; i<GLOBAL_HASHSIZE; ++i)
  {
    struct GlobalElement *v;

    for (v=this->table[i]; v; v=v->next)
    {
      if (v->sym.type==GLOBALVAR) Var_clear(&(v->sym.u.var));
    }
  }
}
/*}}}*/
void Global_destroy(struct Global *this) /*{{{*/
{
  int i;

  for (i=0; i<GLOBAL_HASHSIZE; ++i)
  {
    struct GlobalElement *v=this->table[i],*w;
    struct Symbol *sym;

    while (v)
    {
      sym=&(v->sym);
      w=v->next;
      switch (sym->type)
      {
        case GLOBALVAR: Var_destroy(&(sym->u.var)); break;
        case USERFUNCTION:
        {
          if (sym->u.sub.u.def.localTypes) free(sym->u.sub.u.def.localTypes);
          if (sym->u.sub.argTypes) free(sym->u.sub.argTypes);
          break;
        }
        case BUILTINFUNCTION:
        {
          if (sym->u.sub.argTypes) free(sym->u.sub.argTypes);
          if (sym->u.sub.u.bltin.next)
          {
            sym=sym->u.sub.u.bltin.next;
            while (sym)
            {
              struct Symbol *n;

              if (sym->u.sub.argTypes) free(sym->u.sub.argTypes);
              n=sym->u.sub.u.bltin.next;
              free(sym);
              sym=n;
            }
          }
          break;
        }
        default: assert(0);
      }
      free(v->identifier);
      free(v);
      v=w;
    }
    this->table[i]=(struct GlobalElement*)0;
  }
}
/*}}}*/
